@@ -1,22 +1,26 @@
+# frozen_string_literal: true
+
 require 'grape/router/route'
+require 'grape/util/cache'
 
 module Grape
   class Router
     attr_reader :map, :compiled
 
-    class Any < AttributeTranslator
-      def initialize(pattern, **attributes)
-        @pattern = pattern
-        super(attributes)
+    class NormalizePathCache < Grape::Util::Cache
+      def initialize
+        @cache = Hash.new do |h, path|
+          normalized_path = +"/#{path}"
+          normalized_path.squeeze!('/')
+          normalized_path.sub!(%r{/+\Z}, '')
+          normalized_path = '/' if normalized_path.empty?
+          h[path] = -normalized_path
+        end
       end
     end
 
     def self.normalize_path(path)
-      path = "/#{path}"
-      path.squeeze!('/')
-      path.sub!(%r{/+\Z}, '')
-      path = '/' if path == ''
-      path
+      NormalizePathCache[path]
     end
 
     def self.supported_methods
@@ -25,18 +29,20 @@ module Grape
 
     def initialize
       @neutral_map = []
+      @neutral_regexes = []
       @map = Hash.new { |hash, key| hash[key] = [] }
       @optimized_map = Hash.new { |hash, key| hash[key] = // }
     end
 
     def compile!
       return if compiled
-      @union = Regexp.union(@neutral_map.map(&:regexp))
+      @union = Regexp.union(@neutral_regexes)
+      @neutral_regexes = nil
       self.class.supported_methods.each do |method|
         routes = map[method]
         @optimized_map[method] = routes.map.with_index do |route, index|
           route.index = index
-          route.regexp = /(?<_#{index}>#{route.pattern.to_regexp})/
+          Regexp.new("(?<_#{index}>#{route.pattern.to_regexp})")
         end
         @optimized_map[method] = Regexp.union(@optimized_map[method])
       end
@@ -44,12 +50,12 @@ module Grape
     end
 
     def append(route)
-      map[route.request_method.to_s.upcase] << route
+      map[route.request_method] << route
     end
 
     def associate_routes(pattern, **options)
-      regexp = /(?<_#{@neutral_map.length}>)#{pattern.to_regexp}/
-      @neutral_map << Any.new(pattern, regexp: regexp, index: @neutral_map.length, **options)
+      @neutral_regexes << Regexp.new("(?<_#{@neutral_map.length}>)#{pattern.to_regexp}")
+      @neutral_map << Grape::Router::AttributeTranslator.new(options.merge(pattern: pattern, index: @neutral_map.length))
     end
 
     def call(env)
@@ -101,7 +107,7 @@ module Grape
         env,
         neighbor.allow_header,
         neighbor.endpoint
-      ) if neighbor && method == 'OPTIONS' && !cascade
+      ) if neighbor && method == Grape::Http::Headers::OPTIONS && !cascade
 
       route = match?(input, '*')
       return neighbor.endpoint.call(env) if neighbor && cascade && route
@@ -155,7 +161,7 @@ module Grape
     end
 
     def call_with_allow_headers(env, methods, endpoint)
-      env[Grape::Env::GRAPE_ALLOWED_METHODS] = methods
+      env[Grape::Env::GRAPE_ALLOWED_METHODS] = methods.join(', ').freeze
       endpoint.call(env)
     end
 
