@@ -1,3 +1,5 @@
+# frozen_string_literal: true
+
 require 'spec_helper'
 
 describe Grape::Validations do
@@ -9,14 +11,17 @@ describe Grape::Validations do
 
   describe 'params' do
     context 'optional' do
-      it 'validates when params is present' do
+      before do
         subject.params do
           optional :a_number, regexp: /^[0-9]+$/
+          optional :attachment, type: File
         end
         subject.get '/optional' do
           'optional works!'
         end
+      end
 
+      it 'validates when params is present' do
         get '/optional', a_number: 'string'
         expect(last_response.status).to eq(400)
         expect(last_response.body).to eq('a_number is invalid')
@@ -27,14 +32,7 @@ describe Grape::Validations do
       end
 
       it "doesn't validate when param not present" do
-        subject.params do
-          optional :a_number, regexp: /^[0-9]+$/
-        end
-        subject.get '/optional' do
-          'optional works!'
-        end
-
-        get '/optional'
+        get '/optional', a_number: nil, attachment: nil
         expect(last_response.status).to eq(200)
         expect(last_response.body).to eq('optional works!')
       end
@@ -120,6 +118,62 @@ describe Grape::Validations do
       end
     end
 
+    context 'requires with nested params' do
+      before do
+        subject.params do
+          requires :first_level, type: Hash do
+            optional :second_level, type: Array do
+              requires :value, type: Integer
+              optional :name, type: String
+              optional :third_level, type: Array do
+                requires :value, type: Integer
+                optional :name, type: String
+                optional :fourth_level, type: Array do
+                  requires :value, type: Integer
+                  optional :name, type: String
+                end
+              end
+            end
+          end
+        end
+        subject.put('/required') { 'required works' }
+      end
+
+      let(:request_params) do
+        {
+          first_level: {
+            second_level: [
+              { value: 1, name: 'Lisa' },
+              {
+                value: 2,
+                name: 'James',
+                third_level: [
+                  { value: 'three', name: 'Sophie' },
+                  {
+                    value: 4,
+                    name: 'Jenny',
+                    fourth_level: [
+                      { name: 'Samuel' }, { value: 6, name: 'Jane' }
+                    ]
+                  }
+                ]
+              }
+            ]
+          }
+        }
+      end
+
+      it 'validates correctly in deep nested params' do
+        put '/required', request_params.to_json, 'CONTENT_TYPE' => 'application/json'
+
+        expect(last_response.status).to eq(400)
+        expect(last_response.body).to eq(
+          'first_level[second_level][1][third_level][0][value] is invalid, ' \
+          'first_level[second_level][1][third_level][1][fourth_level][0][value] is missing'
+        )
+      end
+    end
+
     context 'requires :all using Grape::Entity documentation' do
       def define_requires_all
         documentation = {
@@ -436,7 +490,7 @@ describe Grape::Validations do
           class DateRangeValidator < Grape::Validations::Base
             def validate_param!(attr_name, params)
               return if params[attr_name][:from] <= params[attr_name][:to]
-              raise Grape::Exceptions::Validation, params: [@scope.full_name(attr_name)], message: "'from' must be lower or equal to 'to'"
+              raise Grape::Exceptions::Validation.new(params: [@scope.full_name(attr_name)], message: "'from' must be lower or equal to 'to'")
             end
           end
         end
@@ -520,7 +574,7 @@ describe Grape::Validations do
         # NOTE: with body parameters in json or XML or similar this
         # should actually fail with: children[parents][name] is missing.
         expect(last_response.status).to eq(400)
-        expect(last_response.body).to eq('children[1][parents] is missing')
+        expect(last_response.body).to eq('children[1][parents] is missing, children[0][parents][1][name] is missing, children[0][parents][1][name] is empty')
       end
 
       it 'errors when a parameter is not present in array within array' do
@@ -540,7 +594,10 @@ describe Grape::Validations do
         ]
 
         expect(last_response.status).to eq(400)
-        expect(last_response.body).to eq('children[0][parents] is missing, children[1][parents] is missing')
+        expect(last_response.body).to eq(
+          'children[0][parents][0][name] is missing, ' \
+          'children[1][parents][0][name] is missing'
+        )
       end
 
       it 'safely handles empty arrays and blank parameters' do
@@ -548,10 +605,17 @@ describe Grape::Validations do
         # should actually return 200, since an empty array is valid.
         get '/within_array', children: []
         expect(last_response.status).to eq(400)
-        expect(last_response.body).to eq('children is missing')
+        expect(last_response.body).to eq(
+          'children[0][name] is missing, ' \
+          'children[0][parents] is missing, ' \
+          'children[0][parents] is invalid, ' \
+          'children[0][parents][0][name] is missing, ' \
+          'children[0][parents][0][name] is empty'
+        )
+
         get '/within_array', children: [name: 'Jay']
         expect(last_response.status).to eq(400)
-        expect(last_response.body).to eq('children[0][parents] is missing')
+        expect(last_response.body).to eq('children[0][parents] is missing, children[0][parents][0][name] is missing, children[0][parents][0][name] is empty')
       end
 
       it 'errors when param is not an Array' do
@@ -699,7 +763,7 @@ describe Grape::Validations do
         expect(last_response.status).to eq(200)
         put_with_json '/within_array', children: [name: 'Jay']
         expect(last_response.status).to eq(400)
-        expect(last_response.body).to eq('children[0][parents] is missing')
+        expect(last_response.body).to eq('children[0][parents] is missing, children[0][parents][0][name] is missing')
       end
     end
 
@@ -774,7 +838,7 @@ describe Grape::Validations do
       it 'does internal validations if the outer group is present' do
         get '/nested_optional_group', items: [{ key: 'foo' }]
         expect(last_response.status).to eq(400)
-        expect(last_response.body).to eq('items[0][required_subitems] is missing')
+        expect(last_response.body).to eq('items[0][required_subitems] is missing, items[0][required_subitems][0][value] is missing')
 
         get '/nested_optional_group', items: [{ key: 'foo', required_subitems: [{ value: 'bar' }] }]
         expect(last_response.status).to eq(200)
@@ -794,7 +858,7 @@ describe Grape::Validations do
       it 'handles validation within arrays' do
         get '/nested_optional_group', items: [{ key: 'foo' }]
         expect(last_response.status).to eq(400)
-        expect(last_response.body).to eq('items[0][required_subitems] is missing')
+        expect(last_response.body).to eq('items[0][required_subitems] is missing, items[0][required_subitems][0][value] is missing')
 
         get '/nested_optional_group', items: [{ key: 'foo', required_subitems: [{ value: 'bar' }] }]
         expect(last_response.status).to eq(200)
@@ -841,7 +905,7 @@ describe Grape::Validations do
         class Customvalidator < Grape::Validations::Base
           def validate_param!(attr_name, params)
             return if params[attr_name] == 'im custom'
-            raise Grape::Exceptions::Validation, params: [@scope.full_name(attr_name)], message: 'is not custom!'
+            raise Grape::Exceptions::Validation.new(params: [@scope.full_name(attr_name)], message: 'is not custom!')
           end
         end
       end
@@ -989,7 +1053,7 @@ describe Grape::Validations do
           class CustomvalidatorWithOptions < Grape::Validations::Base
             def validate_param!(attr_name, params)
               return if params[attr_name] == @option[:text]
-              raise Grape::Exceptions::Validation, params: [@scope.full_name(attr_name)], message: message
+              raise Grape::Exceptions::Validation.new(params: [@scope.full_name(attr_name)], message: message)
             end
           end
         end
@@ -1013,7 +1077,7 @@ describe Grape::Validations do
           expect(last_response.body).to eq('custom is not custom with options!')
         end
       end
-    end # end custom validation
+    end
 
     context 'named' do
       context 'can be defined' do
@@ -1223,7 +1287,9 @@ describe Grape::Validations do
             end
             get '/custom_message/mutually_exclusive', beer: 'true', wine: 'true', nested: { scotch: 'true', aquavit: 'true' }, nested2: [{ scotch2: 'true' }, { scotch2: 'true', aquavit2: 'true' }]
             expect(last_response.status).to eq(400)
-            expect(last_response.body).to eq 'beer, wine are mutually exclusive pass only one, scotch, aquavit are mutually exclusive pass only one, scotch2, aquavit2 are mutually exclusive pass only one'
+            expect(last_response.body).to eq(
+              'beer, wine are mutually exclusive pass only one, nested[scotch], nested[aquavit] are mutually exclusive pass only one, nested2[1][scotch2], nested2[1][aquavit2] are mutually exclusive pass only one'
+            )
           end
         end
 
@@ -1249,7 +1315,7 @@ describe Grape::Validations do
 
           get '/mutually_exclusive', beer: 'true', wine: 'true', nested: { scotch: 'true', aquavit: 'true' }, nested2: [{ scotch2: 'true' }, { scotch2: 'true', aquavit2: 'true' }]
           expect(last_response.status).to eq(400)
-          expect(last_response.body).to eq 'beer, wine are mutually exclusive, scotch, aquavit are mutually exclusive, scotch2, aquavit2 are mutually exclusive'
+          expect(last_response.body).to eq 'beer, wine are mutually exclusive, nested[scotch], nested[aquavit] are mutually exclusive, nested2[1][scotch2], nested2[1][aquavit2] are mutually exclusive'
         end
       end
 
@@ -1318,7 +1384,7 @@ describe Grape::Validations do
               optional :beer
               optional :wine
               optional :juice
-              exactly_one_of :beer, :wine, :juice, message: { exactly_one: 'are missing, exactly one parameter is required', mutual_exclusion: 'are mutually exclusive, exactly one parameter is required' }
+              exactly_one_of :beer, :wine, :juice, message: 'are missing, exactly one parameter is required'
             end
             get '/exactly_one_of' do
               'exactly_one_of works!'
@@ -1352,7 +1418,7 @@ describe Grape::Validations do
           it 'errors when two or more are present' do
             get '/custom_message/exactly_one_of', beer: 'string', wine: 'anotherstring'
             expect(last_response.status).to eq(400)
-            expect(last_response.body).to eq 'beer, wine are mutually exclusive, exactly one parameter is required'
+            expect(last_response.body).to eq 'beer, wine are missing, exactly one parameter is required'
           end
         end
 
@@ -1399,7 +1465,7 @@ describe Grape::Validations do
         it 'errors when none are present' do
           get '/exactly_one_of_nested'
           expect(last_response.status).to eq(400)
-          expect(last_response.body).to eq 'nested is missing, beer_nested, wine_nested, juice_nested are missing, exactly one parameter must be provided'
+          expect(last_response.body).to eq 'nested is missing, nested[beer_nested], nested[wine_nested], nested[juice_nested] are missing, exactly one parameter must be provided'
         end
 
         it 'succeeds when one is present' do
@@ -1411,7 +1477,7 @@ describe Grape::Validations do
         it 'errors when two or more are present' do
           get '/exactly_one_of_nested', nested: { beer_nested: 'string' }, nested2: [{ beer_nested2: 'string', wine_nested2: 'anotherstring' }]
           expect(last_response.status).to eq(400)
-          expect(last_response.body).to eq 'beer_nested2, wine_nested2 are mutually exclusive'
+          expect(last_response.body).to eq 'nested2[0][beer_nested2], nested2[0][wine_nested2] are mutually exclusive'
         end
       end
     end
@@ -1485,16 +1551,16 @@ describe Grape::Validations do
         before :each do
           subject.params do
             requires :nested, type: Hash do
-              optional :beer_nested
-              optional :wine_nested
-              optional :juice_nested
-              at_least_one_of :beer_nested, :wine_nested, :juice_nested
+              optional :beer
+              optional :wine
+              optional :juice
+              at_least_one_of :beer, :wine, :juice
             end
             optional :nested2, type: Array do
-              optional :beer_nested2
-              optional :wine_nested2
-              optional :juice_nested2
-              at_least_one_of :beer_nested2, :wine_nested2, :juice_nested2
+              optional :beer
+              optional :wine
+              optional :juice
+              at_least_one_of :beer, :wine, :juice
             end
           end
           subject.get '/at_least_one_of_nested' do
@@ -1505,17 +1571,17 @@ describe Grape::Validations do
         it 'errors when none are present' do
           get '/at_least_one_of_nested'
           expect(last_response.status).to eq(400)
-          expect(last_response.body).to eq 'nested is missing, beer_nested, wine_nested, juice_nested are missing, at least one parameter must be provided'
+          expect(last_response.body).to eq 'nested is missing, nested[beer], nested[wine], nested[juice] are missing, at least one parameter must be provided'
         end
 
         it 'does not error when one is present' do
-          get '/at_least_one_of_nested', nested: { beer_nested: 'string' }, nested2: [{ beer_nested2: 'string' }]
+          get '/at_least_one_of_nested', nested: { beer: 'string' }, nested2: [{ beer: 'string' }]
           expect(last_response.status).to eq(200)
           expect(last_response.body).to eq 'at_least_one_of works!'
         end
 
         it 'does not error when two are present' do
-          get '/at_least_one_of_nested', nested: { beer_nested: 'string', wine_nested: 'string' }, nested2: [{ beer_nested2: 'string', wine_nested2: 'string' }]
+          get '/at_least_one_of_nested', nested: { beer: 'string', wine: 'string' }, nested2: [{ beer: 'string', wine: 'string' }]
           expect(last_response.status).to eq(200)
           expect(last_response.body).to eq 'at_least_one_of works!'
         end
